#!/usr/bin/env python
import argparse
import sqlite3
import sys
import urllib.parse as urlparse

RED = "\033[31m"
MAGENTA = "\033[35m"
BLUE = "\033[34m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
LBLUE = "\033[36m"
ENDC = "\033[0m"

colors = {"200": GREEN, "302": BLUE, "403": RED, "500": YELLOW}

response_codes = ["200"]

QUERY = "SELECT * FROM requests ORDER BY resource"


def viewer(project_name):
    try:
        with sqlite3.connect(project_name) as connection:
            cursor = connection.cursor()
            connection.text_factory = str
            cursor.execute(QUERY)
            data = cursor.fetchall()
            tree(data)

    except Exception as e:
        import traceback

        traceback.print_exc()
        print("Error opening database {}".format(project_name))
        print("Details: {}".format(e))
        sys.exit(-1)


def table(data):
    def comparator(a, b):
        if a.count("/") < b.count("/"):
            return -1
        if a.count("/") > b.count("/"):
            return 1
        if a.count("/") == b.count("/"):
            return 0

    # if used_payloads:
    #     output = {}
    #     for k, v in data:
    #         if not output.has_key(k):
    #             output[k] = [v]
    #             continue
    #         output[k].append(v)
    #     for k in output.keys():
    #         print("{}{}:".format(GREEN, k))
    #         output[k].sort()
    #         for v in set(output[k]):
    #             print("        {}{}".format(BLUE, v))
    #     sys.exit()

    data.sort(cmp=comparator, key=lambda x: x[2] + x[3] + x[4])
    for line in data:
        (url, resource, extension, response_code, location, response_size) = (
            line[2],
            line[3],
            line[4],
            line[5],
            line[8],
            line[6],
        )

        if response_code in response_codes:
            url = urlparse.urlparse(url)
            if not location:
                print(
                    "{}{} {:>8}: {}{}{}".format(
                        colors[response_code],
                        response_code,
                        response_size,
                        url.path,
                        resource,
                        extension,
                    )
                )
            else:
                print(
                    "{}{} {:>8}: {}{}{} -> {}".format(
                        colors[response_code],
                        response_code,
                        response_size,
                        url.path,
                        resource,
                        extension,
                        location,
                    )
                )


def tree(data):
    try:
        from asciitree import LeftAligned
    except:
        print("[!] Module 'asciitree' not installed. Try: \"pip install asciitree\"")
        sys.exit(-1)

    tr = LeftAligned()
    tree = {"/": {}}

    def _put_into_dict(root, resource):
        resource = resource
        component = resource.split("/")[0]
        if not component:
            return
        if not component in root.keys():
            root[component] = {}
        resource = "/".join(resource.split("/")[1:])
        _put_into_dict(root[component], resource=resource)

    for entry in data:
        (resource, code, ext, response_size) = (entry[3], entry[5], entry[4], entry[6])
        if code in response_codes:
            if ext:
                resource = resource + ext
            _put_into_dict(tree["/"], resource)
    print(tr(tree))



# -*- coding: utf-8 -*-

import sys
import urlparse
import threading

try:
    import requests
    from BeautifulSoup import BeautifulSoup
except:
    print("[CRAWLER] Oops, check your dependencies are installed")
    print("[CRAWLER] pip install requests bs4")


USER_AGENT = "Mozilla/5.0 (Windows; U; MSIE 10.0; Windows NT 9.0; en-EN)"
user_agent = {"user-agent": USER_AGENT}
DESIRABLE_EXTENSIONS = ['.html', '.htm', '.inc', '.php', '.asp', '.aspx', '.txt']

_non_visited_links = list()
_visited_links = list()
_loot = dict()


def visit(scheme, domain, resource):
    '''
        Visit 'url' with designated 'user_agent'
        return a 'resources' list with unique elements
    '''
    global _non_visited_links
    temporal_resource_list = list()

    if resource.startswith('/'):
        url = "%s://%s%s" % (scheme, domain, resource)
    else:
        url = "%s://%s/%s" % (scheme, domain, resource)

    print("visiting: url: %s" % url)
    request = requests.get(url, headers=user_agent)
    soup = BeautifulSoup(request.text)
    resources = {
        'anchor': (soup.find_all('a'), 'href'),
        'iframe': (soup.find_all('iframe'), 'src'),
        'frame': (soup.find_all('frame'), 'src'),
        'img': (soup.find_all('img'), 'src'),
        'link': (soup.find_all('link'), 'href'),
        'script': (soup.find_all('script'), 'src'),
        'form': (soup.find_all('form'), 'action'),
        }

    for res in resources.values():
        tags, attr = res
        for tag in tags:
            if tag.has_attr(attr):
                temporal_resource_list.append(tag[attr])

    if resource.startswith('/'):
        resource = resource[1:]

    _visited_links.append(resource)
    _non_visited_links.extend(resource_filter(domain, temporal_resource_list))


def resource_filter(domain, resource_set):
    my_non_visited_links = []
    for resource in resource_set:
        p_resource = urlparse.urlparse(resource).path
        p_domain = urlparse.urlparse(resource).netloc
        if p_resource.startswith('/'):
            p_resource = p_resource[1:]
        if p_domain and not p_domain == domain:
            continue
        if p_resource not in _visited_links and p_resource not in _non_visited_links:
            if is_interesting(p_resource):
                my_non_visited_links.append(p_resource)
    return list(set(my_non_visited_links))


def is_interesting(resource):
    '''
        Filter a resource container with non-desired resources
    '''
    if resource == '/' or resource == '':
        return False

    # Remove #resouces or javascript functions
    banned_set = set(['(', ')', '#'])
    check_set = set(resource)
    if any(banned_set & check_set):
        return False

    # Keep only desirable directories and files
    last_path_component = urlparse.urlparse(resource).path.split('/')[-1]
    if '.' not in last_path_component:
        return True
    else:
        for extension in DESIRABLE_EXTENSIONS:
            if extension in last_path_component:
                return True
    return False


def recursive_dict_key_finder(key, dictionary):
    if key in dictionary:
        dictionary = dictionary[key]
        return recursive_dict_key_finder(key, dictionary)
    else:
        dictionary[key] = {}
        return dictionary


def get_into_loot(resource):
    '''
        Take a list of resources, break them into pieces to extract directories
        and files
    '''
    pieces = resource.split('/')
    for piece in pieces:
        recursive_dict_key_finder(piece, _loot)


def check_for_302(url):
    req = requests.get(url)
    if req.status_code in ['301', '302']:
        print("[CRAWLER] Crawling relocation instead -> %s" % req.history[0].url)
        return req.history[0].url
    else:
        return url


if __name__ == '__main__':
    url = check_for_302(sys.argv[1])
    components = urlparse.urlparse(url)
    if not components.path:
        _non_visited_links.append('/')
    else:
        _non_visited_links.append(components.path)
    while len(_non_visited_links) > 0:
        try:
            current_resource = _non_visited_links.pop()
            visit(scheme=components.scheme, domain=components.netloc, resource=current_resource)
        except KeyboardInterrupt:
            print("Was visiting %s://%s%s" % (components.scheme, components.netloc, current_resource))
    print(_visited_links)



#
#   Utility functions
#
import sys
import socket
import urllib.parse as urlparse


def check_domain(target_url):
    """Get the target url from the user, clean and return it"""

    domain = urlparse.urlparse(target_url).hostname

    try:
        if socket.gethostbyname(domain):
            pass
    except Exception:
        print("[!] Domain doesn't resolve. Check URL")
        sys.exit(1)


def prepare_target(target_url):
    """Examine target url compliance adding default handle (http://) and look for a final /"""

    if target_url.startswith("http://") or target_url.startswith("https://"):
        pass
    else:
        target_url = "http://" + target_url
    if target_url.endswith("/") or "***" in target_url:
        pass
    else:
        target_url += "/"
    check_domain(target_url)
    return target_url


def prepare_proxies(proxies):
    """It takes a list of proxies and returns a dictionary"""

    if proxies:
        proxies_dict = {}
        for proxy_item in proxies:
            if proxy_item.startswith("http://"):
                proxies_dict["http"] = proxy_item
            elif proxy_item.startswith("https://"):
                proxies_dict["https"] = proxy_item
        return proxies_dict
    return {}


def make_cookie_jar(cookies):
    d = dict()
    c = cookies.split(",")
    if c[0] == "":
        return d
    for entry in c:
        key, value = entry.split(":")
        d[key] = value
    return d



#!/usr/bin/env python3
import argparse
import os
import pickle
import random
import sys
import threading
import time
import urllib.parse as urlparse
from datetime import timedelta

import urllib3

from cansina.core.dbmanager import DBManager
from cansina.core.payload import Payload
from cansina.core.printer import Console
from cansina.core.resumer import Resumer
from cansina.core.viewer import viewer
from cansina.core.visitor import Visitor, strict_codes
from cansina.plugins.inspector import Inspector
from cansina.plugins.robots import process_robots
from cansina.utils.misc import make_cookie_jar, prepare_proxies, prepare_target

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


current_directory = os.path.dirname(os.path.abspath(__file__))

#   Default options
#
THREADS = 4
USER_AGENT = ""

# Set a random user-agent
with open(f"{current_directory}/utils/ua.txt", "r") as uas:
    lines = uas.readlines()
    USER_AGENT = lines[random.randint(0, len(lines) - 1)].strip()

#
# Parsing program options
#
usage = "cansina.py -u url -p payload [options]"
description = """
Cansina is a web content discovery tool.
It makes requests and analyze the responses trying to figure out whether the
resource is or not accessible.
"""
epilog = "License, requests, etc: https://github.com/deibit/cansina"

parser = argparse.ArgumentParser(usage=usage, description=description, epilog=epilog)
parser.add_argument(
    "-A",
    dest="authentication",
    help="Basic Authentication (e.g: user:password)",
    default=False,
)
parser.add_argument(
    "-C", dest="cookies", help="your cookies (e.g: key:value)", default=""
)
parser.add_argument(
    "-D",
    dest="autodiscriminator",
    help="Check for fake 404 (warning: machine decision)",
    action="store_true",
    default=False,
)
parser.add_argument(
    "-H", dest="request_type", help="Make HTTP HEAD requests", action="store_true"
)
parser.add_argument(
    "-P",
    dest="proxies",
    help="Set a http and/or https proxy (ex: http://127.0.0.1:8080,https://...",
    default="",
)
parser.add_argument(
    "-S",
    dest="remove_slash",
    help="Remove ending slash for payloads",
    default=False,
    action="store_true",
)
parser.add_argument(
    "-T",
    dest="request_delay",
    help="Time (a float number, e.g: 0.25 or 1.75) between requests",
    default=0,
)
parser.add_argument(
    "-U",
    dest="uppercase",
    help="Make payload requests upper-case",
    action="store_true",
    default=False,
)
parser.add_argument(
    "-a",
    dest="user_agent",
    help="The preferred user-agent (default provided)",
    default=USER_AGENT,
)
parser.add_argument(
    "-b", dest="banned", help="List of banned response codes", default="404"
)
parser.add_argument(
    "-B",
    dest="unbanned",
    help="List of unbanned response codes, mark all response as invalid without unbanned response codes, higher priority than banned",
    default="",
)
parser.add_argument(
    "-c",
    dest="content",
    help="Inspect content looking for a particular string",
    default="",
)
parser.add_argument(
    "-d",
    dest="discriminator",
    help="If this string if found it will be treated as a 404",
    default=None,
)
parser.add_argument(
    "-e",
    dest="extension",
    help="Extension list to use e.g: php,asp,...(default none)",
    default="",
)
parser.add_argument(
    "-o",
    dest="output",
    help="Write (append) results in CSV format to a file; -o <filename>",
    default="",
)

parser.add_argument(
    "-O",
    dest="db_output",
    help="Where to write the sqlite database file. -O <filepath> (default is current directory)",
    default=".",
)

parser.add_argument(
    "-p",
    dest="payload",
    help="A single file, a file with filenames (.payload) or a directory (will do *.txt)",
    default=None,
)
parser.add_argument(
    "-s",
    dest="size_discriminator",
    help="Will skip pages with this size in bytes (or a list of sizes 0,500,1500...)",
    default=False,
)
parser.add_argument(
    "-t",
    dest="threads",
    type=int,
    help="Number of threads (default 4)",
    default=THREADS,
)
parser.add_argument(
    "-V",
    dest="view",
    help="Print a tree with a project's discovered items: -V output/project.sqlite",
    default="",
)
parser.add_argument("-u", dest="target", help="Target url", default=None)
parser.add_argument("-r", dest="resume", help="Resume a session", default=False)
parser.add_argument(
    "-R",
    dest="parse_robots",
    action="store_true",
    help="Parse robots.txt and check its contents",
    default=False,
)
parser.add_argument(
    "--recursive",
    dest="recursive",
    help="Recursive descend on path directories",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--no-persist",
    dest="persist",
    help="Do not Use HTTP persistent connections",
    default=True,
    action="store_false",
)
parser.add_argument(
    "--full-path",
    dest="full_path",
    help="Show full path instead of only resources",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--show-type",
    dest="show_content_type",
    help="Show content-type in results",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--no-follow",
    dest="allow_redirects",
    help="Do not follow redirections",
    default=True,
    action="store_false",
)
parser.add_argument(
    "--line",
    dest="continue_line",
    type=int,
    help="Continue payload in line <n>",
    default=0,
)
parser.add_argument(
    "--resumer",
    dest="do_resumer",
    help="Save session in a file",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--headers",
    dest="headers",
    help="Set personalized headers: key=value;key=value...",
    default="",
)
parser.add_argument(
    "--capitalize",
    dest="capitalize",
    help="Transform 'word' into 'Word'.",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--strip-extension",
    dest="strip_extension",
    help="Strip word extension: word.ext into word",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--alpha",
    dest="only_alpha",
    help="Filter non alphanumeric words from wordlist",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--no-progress",
    dest="no_progress",
    help="Don't show tested words and progress. (For dumb terminals)",
    default=False,
    action="store_true",
)
parser.add_argument(
    "--no-colors",
    dest="no_colors",
    help="Don't use output colors to keep output clean, e.g. when redirecting output to file",
    default=False,
    action="store_true",
)

args = parser.parse_args()

# Viewer? Show results and exit
view = args.view
if view:
    viewer(view)
    sys.exit()

# Initialize a Resumer object
resumer = Resumer(args, args.continue_line)
resume = args.resume
# If we are ressuming a former session revive last args object
if resume:
    try:
        with open(resume, "rb") as f:
            resumer = pickle.load(f)
            args = resumer.get_args()
    except Exception:
        sys.stdout.write("[!] Could not load a correct resume file, sorry.")
        sys.exit()

# Target check and preparation
if not args.target:
    print("[!] You need to specify a target")
    parser.print_help()
    sys.exit()
target = prepare_target(args.target)

Console.add_config("{:37} {:>}".format("Target:", target))

recursive = args.recursive
Console.add_config(
    "{:37} {:>}".format("Recursive requests", "Yes" if recursive else "No")
)

# Persistent connections
persist = args.persist
Console.add_config(
    "{:37} {:>}".format("Persistent connections", "Yes" if persist else "No")
)


# Misc options
extension = args.extension.split(",")
threads = int(args.threads)
banned_response_codes = [int(item) for item in args.banned.split(",") if item]
unbanned_response_codes = [
    int(item)
    for item in args.unbanned.split(",")
    if item and item not in banned_response_codes
]
user_agent = args.user_agent
proxy = prepare_proxies(args.proxies.split(","))
cookies = args.cookies
request_delay = args.request_delay
authentication = args.authentication
size_discriminator = args.size_discriminator
output = args.output

# Personalized headers
personalized_headers = {}
try:
    if args.headers:
        for header in args.headers.split(","):
            k, v = header.split("=")
            personalized_headers[k] = v
except Exception as e:
    print("[?] Check personalized headers format: header=value,header=value...")
    sys.exit()

# HEAD / GET requests
request_type = args.request_type
if request_type:
    request_type = "HEAD"
else:
    request_type = "GET"

Console.add_config("{:37} {:>}".format("Requests type:", request_type))

# Content inspecting
content = args.content

# Remove slash (probably an ancient option)
remove_slash = args.remove_slash

# Discriminator option
discriminator = args.discriminator

# Autodiscriminator (probably deprecated by future diagnostic subsystem)
autodiscriminator = args.autodiscriminator
autodiscriminator_location = None
autodiscriminator_md5 = None
if autodiscriminator:
    i = Inspector(target)
    (result, notfound_type) = i.check_this()
    if notfound_type == Inspector.TEST404_URL:
        autodiscriminator_location = result
    elif notfound_type == Inspector.TEST404_MD5:
        autodiscriminator_md5 = result

# Misc user information


Console.add_config(
    "{:37} {:>}".format(
        "Filtered response codes:",
        ",".join([str(item) for item in banned_response_codes]),
    )
)


unbanned_response_codes = (
    unbanned_response_codes if unbanned_response_codes else strict_codes
)
unbanned_codes = [str(i) for i in unbanned_response_codes]
Console.add_config("{:37} {:>}".format("Ok response codes:", ",".join(unbanned_codes)))

# Using Custom File Extension List
Console.add_config(
    "{:37} {:>}".format(
        "Custom extensions:", ",".join(extension) if not extension == [""] else "No"
    )
)

# Payload options
# FIXME: This design is garbage
payload = None
parse_robots = args.parse_robots
remove_slash = True
if parse_robots:
    robots_content = process_robots(target)
    if not robots_content:
        print("[!] robots.txt not found")
        sys.exit()
    payload_filename = robots_content
else:
    payload_filename = args.payload
    if not payload_filename:
        print("[!] You have to specify a payload")
        parser.print_help()
        sys.exit()

Console.add_config("{:37} {:>}".format("Payload:", payload_filename))
payload = Payload(target, payload_filename, resumer)

# Uppercase
if args.uppercase:
    payload.set_uppercase()

# Capitalization
if args.capitalize:
    payload.set_capitalize()

# Strip extension
if args.strip_extension:
    payload.set_strip_extension()

# Only alphanumeric words
if args.only_alpha:
    payload.set_alpha()

#
# Payload queue configuration
#
payload.set_extensions(extension)
payload.set_remove_slash(remove_slash)
payload.set_banned_response_codes(banned_response_codes)
payload.set_unbanned_response_codes(unbanned_response_codes)
payload.set_content(content)
payload.set_recursive(recursive)

payload_queue = payload.get_queue()
total_requests = payload.get_total_requests()

#
# Manager queue configuration
#
database_name = (
    urlparse.urlparse(target).scheme + "_" + urlparse.urlparse(target).hostname
)
if urlparse.urlparse(target).port is not None:
    database_name += "_" + str(urlparse.urlparse(target).port)


if args.db_output:
    db_dir = args.db_output
else:
    db_dir = current_directory

manager = DBManager(f"{db_dir}/{database_name}.sqlite")
manager.output = output
manager_lock = threading.Lock()

#
# Configure Visitor Objects
#
Visitor.set_authentication(authentication)
Visitor.set_banned_location(autodiscriminator_location)
Visitor.set_banned_md5(autodiscriminator_md5)
Visitor.set_delay(request_delay)
Visitor.set_discriminator(discriminator)
Visitor.set_proxy(proxy)
Visitor.set_requests(request_type)
Visitor.set_size_discriminator(size_discriminator)
Visitor.set_user_agent(user_agent)
Visitor.set_persist(persist)
Visitor.allow_redirects(args.allow_redirects)
Visitor.set_headers(personalized_headers)

# Cookies
try:
    cookie_jar = make_cookie_jar(cookies)
    Visitor.set_cookies(cookie_jar)
except Exception:
    print("[!] Error setting cookies. Review cookie string (key:value,key:value...)")
    sys.exit()

# Select if full path is prefered
full_path = args.full_path

# Select if user wants content-type print
show_content_type = args.show_content_type

#
#   Run the main thread until manager exhaust all tasks
#
time_before_running = time.time()
Console.number_of_requests = payload.get_total_requests()
Console.number_of_threads = threads
Console.show_full_path = full_path
Console.show_content_type = show_content_type
Console.set_show_progress(False if args.no_progress else True)
Console.set_show_colors(False if args.no_colors else True)
Console.init()

#
# Create the thread_pool and start the daemonized threads
#
thread_pool = []
for visitor_id in range(threads):
    v = Visitor(visitor_id, payload_queue, manager, manager_lock)
    thread_pool.append(v)
    v.start()


if __name__ == "__main__":
    # Main loop
    try:
        payload_queue.join()

        for thread in thread_pool:
            payload_queue.put(None)
        for visitor in thread_pool:
            visitor.join()

    except KeyboardInterrupt:
        Console.say("Waiting for threads to stop...")

        Visitor.kill()

        for visitor in thread_pool:
            visitor.join()

        if args.do_resumer:
            resumer.set_line(payload_queue.get().get_number())
            with open(
                "resume_file_" + time.strftime("%d_%m_%y_%H_%M", time.localtime()), "wb"
            ) as f:
                pickle.dump(resumer, f)

    except Exception as e:
        import traceback as tb

        Console.say(tb.print_tb(sys.exc_info()[2]))

    finally:
        # Dump results to database
        manager.save()
        Console.end()

sys.exit()



import hashlib
import sys
import threading
import time

import requests

unuseful_codes = [404]
strict_codes = [100, 200, 300, 301, 302, 401, 403, 405, 500]


class Visitor(threading.Thread):
    auth = None
    banned_location = None
    banned_md5 = None
    cookies = None
    delay = None
    discriminator = None
    headers = {}
    is_allow_redirects = False
    persist = False
    proxy = None
    requests = ""
    size_discriminator = []
    user_agent = None
    killed = False

    @staticmethod
    def set_headers(headers):
        Visitor.headers = headers

    @staticmethod
    def allow_redirects(pref):
        Visitor.is_allow_redirects = pref

    @staticmethod
    def set_discriminator(discriminator):
        Visitor.discriminator = discriminator

    @staticmethod
    def set_cookies(_cookies):
        Visitor.cookies = _cookies

    @staticmethod
    def kill():
        Visitor.killed = True

    @staticmethod
    def set_size_discriminator(size_discriminator):
        if size_discriminator:
            Visitor.size_discriminator = [int(x) for x in size_discriminator.split(",")]
        else:
            Visitor.size_discriminator = []

    @staticmethod
    def set_banned_location(banned_location):
        Visitor.banned_location = banned_location

    @staticmethod
    def set_banned_md5(banned_md5):
        Visitor.banned_md5 = banned_md5

    @staticmethod
    def set_user_agent(useragent):
        Visitor.user_agent = useragent

    @staticmethod
    def set_proxy(proxy):
        Visitor.proxy = proxy

    @staticmethod
    def set_delay(delay):
        Visitor.delay = float(delay)

    @staticmethod
    def set_requests(type_request):
        Visitor.requests = type_request

    @staticmethod
    def set_authentication(auth):
        Visitor.auth = tuple(auth.split(":")) if auth else auth

    @staticmethod
    def set_persist(persist):
        Visitor.persist = persist

    def __init__(self, visitor_id, payload, results, lock):
        threading.Thread.__init__(self)
        self.visitor_id = visitor_id
        self.payload = payload
        self.results = results
        self.session = None
        self.lock = lock

    def run(self):
        try:
            while not Visitor.killed:
                task = self.payload.get()
                if not task:
                    return
                self.visit(task)
                self.payload.task_done()
        except Exception:
            pass

    def visit(self, task):
        def _dumb_redirect(url):
            origin = "{0}{1}".format(task.target, task.resource)

            # Detect redirect to same page but ended with slash
            if url == origin:
                return True
            if url == origin + "/":
                return True

            # Detect redirect to root
            if url == task.target:
                return True

            return False

        try:
            if Visitor.user_agent:
                Visitor.headers["User-Agent"] = Visitor.user_agent

            # Persistent connections
            if Visitor.persist:
                if not self.session:
                    self.session = requests.Session()
            else:
                self.session = requests

            r = None
            if Visitor.proxy:
                if Visitor.requests == "GET":
                    r = self.session.get(
                        task.get_complete_target(),
                        headers=Visitor.headers,
                        proxies=Visitor.proxy,
                        verify=False,
                        auth=Visitor.auth,
                        cookies=Visitor.cookies,
                        allow_redirects=Visitor.is_allow_redirects,
                    )

                elif Visitor.requests == "HEAD":
                    r = self.session.head(
                        task.get_complete_target(),
                        headers=Visitor.headers,
                        proxies=Visitor.proxy,
                        verify=False,
                        auth=Visitor.auth,
                        cookies=Visitor.cookies,
                        allow_redirects=Visitor.is_allow_redirects,
                    )
            else:
                if Visitor.requests == "GET":
                    r = self.session.get(
                        task.get_complete_target(),
                        headers=Visitor.headers,
                        verify=False,
                        auth=Visitor.auth,
                        cookies=Visitor.cookies,
                        allow_redirects=Visitor.is_allow_redirects,
                    )

                elif Visitor.requests == "HEAD":
                    r = self.session.head(
                        task.get_complete_target(),
                        headers=Visitor.headers,
                        verify=False,
                        auth=Visitor.auth,
                        cookies=Visitor.cookies,
                        allow_redirects=Visitor.is_allow_redirects,
                    )

            tmp_content = r.content
            task.response_size = len(tmp_content)
            task.response_time = round(r.elapsed.microseconds / 1000, 2)

            task.set_response_code(r.status_code)

            # If discriminator is found we mark it 404
            if sys.version_info[0] >= 3:
                tmp_content = tmp_content.decode("Latin-1")
            if Visitor.discriminator and Visitor.discriminator in tmp_content:
                task.ignorable = True

            if (
                Visitor.banned_md5
                and hashlib.md5("".join(tmp_content)).hexdigest() == self.banned_md5
            ):
                task.ignorable = True

            # Check if page size is not what we need
            if task.response_size in Visitor.size_discriminator:
                task.ignorable = True

            # Look for interesting content
            if task.content and (task.content in tmp_content):
                task.content_has_detected(True)

            # Look for a redirection
            if Visitor.is_allow_redirects:
                if len(r.history) > 0 and not _dumb_redirect(r.history[-1].url):
                    task.response_code = str(r.history[0].status_code)
                    task.location = r.history[-1].url
            else:
                if r.status_code >= 300 and r.status_code < 400:
                    task.set_response_code(404)
                    task.ignorable = True

            if "content-type" in [h.lower() for h in r.headers.keys()]:
                try:
                    task.response_type = r.headers["Content-Type"].split(";")[0]
                except Exception:
                    pass

            task.thread = self.visitor_id
            self.lock.acquire()
            self.results.put(task)

            if Visitor.delay:
                time.sleep(Visitor.delay)

        except (requests.ConnectionError, requests.Timeout) as e:
            # TODO log to a file instead of screen
            print("[!] Timeout/Connection error")
            print(e)

        except Exception as e:
            print("[!] General exception while visiting")
            print(e)

        finally:
            self.lock.release()
            return



class Resumer:
    def __init__(self, args, line):
        self.args = args
        self.line = line

    def set_line(self, l):
        self.line = l

    def get_line(self):
        return self.line

    def get_args(self):
        return self.args



import os
import queue
import sqlite3
import sys
import time

from cansina.core.printer import Console

SUFIX = ".sqlite"


class DBManager:
    def __init__(self, database_path):

        # Check for self.database_path, create dirs and database if they don't exists
        self.database_path = database_path
        if not os.path.isfile(self.database_path):
            try:
                connection = sqlite3.connect(self.database_path)
                connection.text_factory = str
                cursor = connection.cursor()
                cursor.execute(
                    "CREATE TABLE requests (\
                                line_number INTEGER, \
                                payload TEXT,\
                                url TEXT,\
                                resource TEXT,\
                                extension TEXT,\
                                response_code TEXT,\
                                response_size INTEGER,\
                                response_time INTEGER,\
                                location TEXT, \
                                t_stamp INTEGER);"
                )
                connection.commit()
                connection.close()

            except Exception as e:
                print("[DBManager] Error creating database {0}".format(database_path))
                sys.exit()

        self.dead = False
        self.queue = queue.Queue()
        self.output = None

    def put(self, task):
        Console.update(task)
        self.queue.put(task)

    def save(self):
        output = None
        if self.output:
            output = open(self.output, "a")

        while not self.queue.empty():
            task = self.queue.get()
            self.process(task)
            if output and not task.ignorable:
                output.write(
                    "{},{},{}\n".format(
                        task.response_code,
                        task.get_complete_target(),
                        task.response_size,
                    )
                )
            self.queue.task_done()

    def process(self, task):
        connection = sqlite3.connect(self.database_path)
        cursor = connection.cursor()
        connection.text_factory = str
        # Check if the record already exists
        record = {
            "url": task.target,
            "resource": task.resource,
            "extension": task.extension,
            "response_code": task.response_code,
        }
        try:
            cursor.execute(
                "SELECT * FROM requests WHERE \
                            url=:url AND \
                            resource=:resource AND \
                            extension=:extension AND \
                            response_code=:response_code",
                record,
            )
        except Exception as e:
            print(e)
        # TODO banned response code at user will
        if not cursor.fetchone():
            if not task.response_code == 404 or not task.ignorable:
                cursor.execute(
                    "INSERT INTO requests VALUES (?,?,?,?,?,?,?,?,?,?)",
                    task.values() + (time.time(),),
                )
                connection.commit()
        connection.close()



import fileinput
import fnmatch
import os
import os.path
import queue
import sys
import threading
import time
import urllib.parse as urlparse

from cansina.core.task import Task


def _get_url_components(target):
    """Get a url and returns multiple paths composed by recursive components"""
    path = [i for i in urlparse.urlparse(target).path.split("/") if len(i) > 0]
    temp_path = ""
    list_path = ["/"]
    for component in path:
        sub = "/%s" % component
        temp_path += sub
        list_path.append(temp_path + "/")
    return list_path


def _populate_list_with_file(file_name, linenumber):
    """Open a file, read its content and strips it. Returns a list with the content
    additionally it filter and clean some splinters
    """

    def _read_a_file_return_a_list(file_name):
        """
        Small utility function. Open a file a return its content as a list
        """
        tmp_list = []
        try:
            if sys.version_info > (3, 0):
                f = open(file_name, "r", encoding="latin-1")
            else:
                f = open(file_name, "r")
            tmp_list = f.readlines()

        except (OSError, IOError) as e:
            print(
                "[!] Opening payload. Check file, list of files or directory content."
            )
            print(e)
            sys.exit()

        return tmp_list

    # FIXME: Garbage
    tmp_list = None
    # If we receive a list then they are robots.txt entries
    if type(file_name) == list:
        tmp_list = file_name
    # If filename is - we are dealing with stdin
    elif file_name == "-":
        tmp_list = []
        for line in fileinput.input(files=file_name):
            tmp_list.append(line)
    # Single file, could be a list of lists or a single payload
    elif os.path.isfile(file_name):
        tmp_list = _read_a_file_return_a_list(file_name)
        # Now is time to check if is a list or single payload
        if not file_name.endswith(".payload"):
            tmp_list = tmp_list[linenumber:]
        else:
            payload_list = tmp_list
            tmp_list = []
            print("[*] Trying to load {0} files as payloads".format(len(payload_list)))
            for item in payload_list:
                tmp_list.extend(_read_a_file_return_a_list(item.strip()))

    # If we receive a dir then try to open all .txt files in there
    elif os.path.isdir(file_name):
        tmp_list = []
        print(
            "[*] Directory payload inclusion. All *.txt will be treated as a payload."
        )
        for file in os.listdir(file_name):
            if fnmatch.fnmatch(file, "*.txt"):
                tmp_list.extend(
                    _read_a_file_return_a_list("{0}/{1}".format(file_name, file))
                )

    clean_list = []
    for e in tmp_list:
        # Delete leading and trailing spaces
        e = e.strip()
        # Skip commented lines in payload files
        if e.startswith("#"):
            continue
        # Remove leading '/' characters
        if e.startswith("/"):
            e = e[1:]

        if sys.version_info[0] == 3:
            e_encode = e
        else:
            e_encode = e.decode("utf-8", "replace")
        clean_list.append(e_encode)
    return clean_list


def _has_extension(res):
    # whether the last path sector has '.'
    if res.rfind("/") == -1:
        return "." in res
    else:
        return "." in res[res.rfind("/") :]


class Payload:
    def __init__(self, target, payload_filename, resumer):

        self.target = target
        self.task_id = 0
        # Payload may be a filename, a python list, a directory or a file with
        # paths to multiple payload files
        self.payload_filename = (
            payload_filename if not type(payload_filename) == list else "robots.txt"
        )
        self.linenumber = resumer.get_line()
        self.payload = _populate_list_with_file(payload_filename, self.linenumber)
        self.queue = queue.Queue()
        self.dead = False
        self.extensions = None
        self.banned_response_codes = None
        self.unbanned_response_codes = None
        self.content = None
        self.remove_slash = False
        self.uppercase = False
        self.capitalize = False
        self.recursive = False
        self.strip_extension = False
        self.only_alpha = False
        self.total_requests = 0
        self.number_of_tasks = 0

    def get_total_requests(self):
        return self.number_of_tasks

    def set_recursive(self, recursion):
        self.recursive = recursion

    def set_remove_slash(self, remove_slash):
        self.remove_slash = remove_slash

    def set_banned_response_codes(self, banned_response_codes):
        self.banned_response_codes = banned_response_codes

    def set_unbanned_response_codes(self, unbanned_response_codes):
        self.unbanned_response_codes = unbanned_response_codes

    def set_extensions(self, extensions):
        self.extensions = extensions

    def set_content(self, content):
        self.content = content

    def kill(self):
        self.dead = True

    def is_finished(self):
        return self.dead

    def set_uppercase(self):
        self.uppercase = True

    def set_capitalize(self):
        self.capitalize = True

    def set_strip_extension(self):
        self.strip_extension = True

    def set_alpha(self):
        self.only_alpha = True

    def _feed_queue(self):
        for resource in self.payload:
            if self.only_alpha:
                if not resource.isalnum():
                    continue

            if self.uppercase:
                resource = resource.upper()

            if self.capitalize:
                resource = resource.capitalize()

            if self.strip_extension:
                resource = resource.split(".")[0]

            # Useful when looking for files without extension instead of directories
            if self.remove_slash and resource.endswith("/"):
                resource = resource[:-1]

            for extension in self.extensions:
                # If resource is a whole word and user didnt provide a extension
                # put a final /
                if (
                    not extension
                    and not _has_extension(resource)
                    and not self.remove_slash
                ):
                    resource += "/"

                # Put a . before extension if the users didnt do it
                if extension and not "." in extension:
                    extension = "." + extension

                self.task_id += 1
                task = Task(self.task_id, self.target, resource, extension)
                task.set_payload_filename(self.payload_filename)
                task.set_payload_length(self.get_total_requests())
                task.set_banned_response_codes(self.banned_response_codes)
                task.set_unbanned_response_codes(self.unbanned_response_codes)
                task.set_content(self.content)
                self.queue.put(task)
                self.number_of_tasks += 1

    def get_queue(self):
        if self.recursive:
            url_components = _get_url_components(self.target)
            # save main component (scheme + netloc)
            path = urlparse.urlparse(self.target).path
            saved_main_component = self.target[: self.target.find(path)]
            self.total_requests = (
                self.number_of_tasks * len(self.extensions) * len(url_components)
            )

            # if user select recursive but url is just root
            if len(url_components) == 1:
                self._feed_queue()
            else:
                # iterate over generated multiple paths
                for i in url_components:
                    self.target = saved_main_component + i
                    self._feed_queue()
        else:
            self.total_requests = self.number_of_tasks * len(self.extensions)
            self._feed_queue()
        return self.queue



import requests
import sys

def process_robots(target):
    '''
        Return a list of entries if robots.txt is found and has meat otherwise
        return None
    '''
    interesting_entries = []
    try:
        result = requests.get(target + "robots.txt")
        if not result.status_code == 200:
            return
        for line in result.text.splitlines():
            if line.startswith("Disallow:") or line.startswith("Allow:"):
                if len(line.split(":")) > 1:
                    interesting_entries.append(line.split(":")[1])
        return list(set(interesting_entries))
    except Exception as e:
        sys.stderr.write("[robots] Error getting robots.txt")
        sys.stderr.write(e)
        return None



from setuptools import setup, find_packages

from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="cansina",
    version="0.9",
    description="Web Content Discovery Tool",
    author="David Garcia",
    author_email="daganu@gmail.com",
    url="https://github.com/deibit/cansina",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "cansina=cansina.main:__main__",
        ],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["requests", "asciitree"],
    package_data={"cansina.utils": ["*.txt"]},
    keywords=["pentesting", "cybersecurity", "websecurity"],
)



import os
import sys
import urllib.parse as urlparse
from collections import OrderedDict

if os.name == "nt":
    RED, MAGENTA, BLUE, GREEN, YELLOW, LBLUE, ENDC = ("", "", "", "", "", "", "")
else:
    RED = "\033[31m"
    MAGENTA = "\033[35m"
    BLUE = "\033[34m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    LBLUE = "\033[36m"
    ENDC = "\033[0m"
    ESC = chr(27)
    GRAY = "\033[38;5;240m"
    CLEARSCR = "\33[2J"
    CURPOS = "\33[{};{}H"
    HIDECUR = "\33[?25l"
    SHOWCUR = "\33[?25h"
    DEL = "\33[2K"

COLUMNS = 80
ROWS = 70


def _get_terminal_width():
    """
    Get the terminal width to adjust columns size
    """
    try:
        global ROWS
        global COLUMNS
        # (Slightly modified) http://stackoverflow.com/a/943921/91267
        p = os.popen("stty size", "r")
        rows, columns = p.read().split()
        p.close()
        COLUMNS = int(columns)
        ROWS = int(rows)

    except Exception:
        # ʕノ•ᴥ•ʔノ ︵ ┻━┻
        # Assume 80 columns and 70 rows at least
        pass


_get_terminal_width()


class Juicy(OrderedDict):
    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)


class Throughput:
    def __init__(self):
        self.MAX = 10
        self.deltas = []

    def push(self, delta):
        if len(self.deltas) == self.MAX:
            self.deltas.pop(0)

        self.deltas.append(delta)

    def get(self):
        deltas_len = len(self.deltas)
        if deltas_len:
            return round(sum(self.deltas) / deltas_len, 2)
        else:
            return 0.0


class Console:
    config_entries = []
    # fixed_height is a value comming from cansina.py configuration info plus banner
    fixed_height = 10 + 5
    juicy_entries = Juicy()
    last_offset = 0
    number_of_requests = 0
    number_of_threads = 0
    show_colors = True
    show_content_type = False
    show_full_path = False
    show_progress = True
    update_counter = 0
    net_throughput = Throughput()

    @staticmethod
    def curpos(x, y):
        sys.stdout.write(CURPOS.format(y, x))

    @staticmethod
    def clear():
        if Console.show_progress:
            sys.stdout.write(CLEARSCR)

    @staticmethod
    def show_cur():
        if Console.show_progress:
            Console.curpos(0, Console.last_offset)
        sys.stdout.write(SHOWCUR)

    @staticmethod
    def hide_cur():
        sys.stdout.write(HIDECUR)

    @staticmethod
    def set_show_progress(show_progress):
        Console.show_progress = show_progress

    @staticmethod
    def set_show_colors(show_colors):
        Console.show_colors = show_colors

    @staticmethod
    def add_config(message):
        Console.config_entries.append(f"{message}\n")

    @staticmethod
    def init():
        if Console.show_progress:
            Console.clear()
            Console.hide_cur()

    @staticmethod
    def end():
        if Console.show_progress:
            Console.show_cur()

    @staticmethod
    def banner():
        ascii_banner = """
   ___              _
  / __|__ _ _ _  __(_)_ _  __ _
 | (__/ _` | ' \(_-< | ' \/ _` |
  \___\__,_|_||_/__/_|_||_\__,_|

        """
        sys.stdout.write(ascii_banner)

    @staticmethod
    def config():
        for line in Console.config_entries:
            sys.stdout.write(line)

    @staticmethod
    def progress():
        block = "▇"
        placeholder = "_"
        number_of_blocks = 60
        block_value = 100 / 60

        percentage = Console.update_counter * 100 / Console.number_of_requests
        blocks_to_print = round(percentage / block_value) + 1
        placeholders_to_print = number_of_blocks - blocks_to_print

        sys.stdout.write(
            f"\r{DEL}{Console.update_counter:>{len(str(Console.number_of_requests))}}/{Console.number_of_requests} [{YELLOW}{block*blocks_to_print}{ENDC}{placeholder*placeholders_to_print}] - {round(percentage,1)}%"
        )

    @staticmethod
    def elapsed_time():
        net_throughput = Console.net_throughput.get()
        color = ""

        if Console.show_colors:
            if net_throughput < 553:
                color = GREEN
            elif net_throughput >= 553 and net_throughput < 998:
                color = YELLOW
            elif net_throughput >= 998:
                color = RED

        sys.stdout.write(
            "\r{}{:37} {}{:>} ms{}\n".format(
                DEL, "Thread activity", color, net_throughput, ENDC
            )
        )

    @staticmethod
    def thread_activity(task):
        color = ""
        if Console.show_colors:
            if task.response_code == 200:
                color = GREEN
            if task.response_code == 401 or task.response_code == 403:
                color = RED
            if task.response_code == 404 or (
                task.response_code in task.banned_response_codes
            ):
                color = GRAY
            if task.response_code == 301 or task.response_code == 302:
                color = LBLUE
            if str(task.response_code).startswith("5") or task.response_code == 400:
                color = YELLOW
            if task.content_detected:
                color = MAGENTA

        target = task.get_complete_target()
        target = urlparse.urlsplit(target).path
        if task.location:
            target = target + " -> " + task.location

        # Full path
        if Console.show_full_path:
            target = task.get_complete_target()

        # Content type
        if Console.show_content_type:
            content_type_color = GRAY if Console.show_colors else ""
            content_type = f"{content_type_color}{task.response_type}{ENDC}"
        else:
            content_type = ""

        # Cut target string if its length is wider than available columns
        if len(target) > COLUMNS:
            target = target[: len(target) - COLUMNS]

        # Format and print info to terminal
        thread_info = f"{color} #{task.thread + 1:<3} {task.response_code}  {target}"
        if Console.show_progress:
            Console.net_throughput.push(task.response_time)
            sys.stdout.write(f"\r{DEL}{thread_info}{ENDC}\n")

        # Add to juicy
        if not task.ignorable:
            size_color = BLUE if Console.show_colors else ""
            formatted_task = f"\r{DEL}{color}{task.response_code:^3}{ENDC} {size_color}{task.response_size:>10} bytes{ENDC} {content_type} {target}"

            if not target in Console.juicy_entries:
                Console.juicy_entries[target] = formatted_task

            if not Console.show_progress:
                formatted_task = f"{color}{task.response_code:^3}{ENDC} {task.response_size:>10} bytes {content_type} {target}\n"
                sys.stdout.write(formatted_task)

    @staticmethod
    def update(task):
        Console.update_counter += 1
        # y-positioning cursor
        # cursor_y = Console.fixed_height
        if Console.show_progress:

            # Show banner
            cursor_y = 0
            Console.curpos(0, cursor_y)
            Console.banner()
            cursor_y += 7

            # Show config
            Console.curpos(0, cursor_y)
            Console.config()
            cursor_y += 9

            # Show thread activity
            cursor_y += 1
            Console.curpos(0, cursor_y)
            Console.elapsed_time()
            cursor_y += 1
            Console.curpos(0, cursor_y + (task.thread + 1))
            Console.thread_activity(task)
            cursor_y += Console.number_of_threads + 1

            # Show progress
            cursor_y += 1
            Console.curpos(0, cursor_y)
            Console.progress()
            cursor_y += 2

            # Show juicy
            juicy_entries = list(Console.juicy_entries.values())
            free_rows = ROWS - cursor_y

            if len(juicy_entries) > free_rows:
                juicy_entries = juicy_entries[len(juicy_entries) - free_rows :]

            for entry in juicy_entries:
                Console.curpos(0, cursor_y)
                cursor_y += 1
                if cursor_y > ROWS:
                    return
                sys.stdout.write(entry)

            Console.last_offset = cursor_y

        else:
            Console.thread_activity(task)

    @staticmethod
    def say(message):
        if Console.show_progress and Console.last_offset < ROWS:
            Console.curpos(0, Console.last_offset)
        sys.stdout.write(f"{message}\n")



class Task:

    """ This class stores information retrieved from/to the request"""

    def __init__(self, number, target, resource, extension):

        self.number = number
        self.target = target
        self.resource = resource
        self.extension = extension
        self.thread = -1

        self.payload_length = 0
        self.payload_filename = None
        self.banned_response_codes = []
        self.unbanned_response_codes = []

        self.content = None
        self.location = ""
        self.response_code = None
        self.response_size = None
        self.response_time = None
        self.response_type = ""
        self.content_detected = False
        self.ignorable = True

    def set_payload_length(self, length):
        self.payload_length = length

    def get_payload_length(self):
        return self.payload_length

    def set_banned_response_codes(self, banned_codes):
        self.banned_response_codes = banned_codes

    def set_unbanned_response_codes(self, unbanned_codes):
        self.unbanned_response_codes = unbanned_codes

    def set_payload_filename(self, payload_filename):
        self.payload_filename = payload_filename

    def set_response_code(self, code):
        self.response_code = code
        # By default a code is valid but...if...
        # ...we select and optimistic detection 'all is valid' unless...
        if self.response_code in self.banned_response_codes:
            self.ignorable = True
        # ...but if we select a pesimistic one 'all is invalid' unless...
        elif self.response_code in self.unbanned_response_codes:
            self.ignorable = False

    def set_location(self, location):
        self.location = location

    def set_content(self, content):
        self.content = content

    def get_content(self):
        return self.content

    def get_number(self):
        return self.number

    def values(self):
        return (
            self.number,
            self.payload_filename,
            self.target,
            self.resource,
            self.extension,
            self.response_code,
            self.response_size,
            self.response_time,
            self.location,
        )

    def get_complete_target(self):
        if "***" in self.target:
            self.target = self.target.replace("***", self.resource.replace("/", ""))
            self.resource = ""
            return self.target + self.extension
        else:
            return self.target + self.resource + self.extension

    def content_has_detected(self, value):
        self.content_detected = value



#!/usr/bin/env python
import sys
import sqlite3
import argparse
import urllib.parse as urlparse

try:
    from asciitree import LeftAligned
except:
    print("install asciitree for -t option")


RED = "\033[31m"
MAGENTA = "\033[35m"
BLUE = "\033[34m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
LBLUE = "\033[36m"
ENDC = "\033[0m"

colors = {"200": GREEN, "302": BLUE, "403": RED, "500": YELLOW}


parser = argparse.ArgumentParser()
parser.add_argument(
    "-p", dest="project_name", help="path to sqlite project database", required=True
)

parser.add_argument(
    "-c",
    dest="response_codes",
    help="comma separate list of http codes responses to print",
    required=False,
)

parser.add_argument(
    "-u", help="list the payload used on visited resources", action="store_true"
)

parser.add_argument(
    "-t",
    dest="tree",
    help="print a tree with the discoreved structure",
    required=False,
    action="store_true",
)

parser.add_argument(
    "-e",
    dest="banned_extensions",
    help="comma separate list of extensions you do not want to see",
    required=False,
)

parser.add_argument(
    "-E",
    dest="unbanned_extensions",
    help="comma separate list of the extensions you only want to see (will include directories)",
    required=False,
)

parser.add_argument(
    "-s",
    dest="size_filter",
    help="comma separate list of size files you do not want to see",
    required=False,
)


args = parser.parse_args()

project_name = args.project_name
response_codes = args.response_codes
banned_extensions = args.banned_extensions
unbanned_extensions = args.unbanned_extensions
size_filter = args.size_filter
used_payloads = args.u
want_tree = args.tree

if not response_codes:
    response_codes = ["200", "302", "403"]

QUERY = "SELECT * FROM requests ORDER BY resource"
if used_payloads:
    QUERY = "SELECT url,payload FROM requests ORDER BY url"

try:
    with sqlite3.connect(project_name) as connection:
        cursor = connection.cursor()
        connection.text_factory = str
        cursor.execute(QUERY)
        data = cursor.fetchall()

except Exception as e:
    print("Error opening database {}".format(project_name))
    print("Details: {}".format(e))
    sys.exit(-1)


def is_banned(resource):
    for banned in banned_extensions.split(","):
        if banned in resource.split("/")[-1]:
            return True
    return False


def is_ubanned(resource):
    if not "." in resource.split("/")[-1]:
        return True
    for unbanned in unbanned_extensions.split(","):
        if unbanned in resource.split("/")[-1]:
            return True
    return False


def size_is_banned(size):
    if str(size) in size_filter.split(","):
        return True
    return False


def table():
    def comparator(a, b):
        if a.count("/") < b.count("/"):
            return -1
        if a.count("/") > b.count("/"):
            return 1
        if a.count("/") == b.count("/"):
            return 0

    if used_payloads:
        output = {}
        for k, v in data:
            if not k in output:
                output[k] = [v]
                continue
            output[k].append(v)
        for k in output.keys():
            print("{}{}:".format(GREEN, k))
            output[k].sort()
            for v in set(output[k]):
                print("        {}{}".format(BLUE, v))
        sys.exit()

    data.sort(key=lambda x: x[2] + x[3] + x[4])
    for line in data:
        (url, resource, extension, response_code, location, response_size) = (
            line[2],
            line[3],
            line[4],
            line[5],
            line[8],
            line[6],
        )

        if banned_extensions and is_banned(resource):
            continue

        if unbanned_extensions and not is_ubanned(resource):
            continue

        if size_filter and size_is_banned(response_size):
            continue

        if response_code in response_codes:
            url = urlparse.urlparse(url)
            if not location:
                print(
                    "{}{} {:>8}: {}{}{}".format(
                        colors[response_code],
                        response_code,
                        response_size,
                        url.path,
                        resource,
                        extension,
                    )
                )
            else:
                print(
                    "{}{} {:>8}: {}{}{} -> {}".format(
                        colors[response_code],
                        response_code,
                        response_size,
                        url.path,
                        resource,
                        extension,
                        location,
                    )
                )


def ftree():
    if not LeftAligned:
        print("asciitree not installed!")
        sys.exit(-1)

    tr = LeftAligned()
    tree = {"/": {}}

    def _put_into_dict(root, resource):
        resource = resource.decode("UTF-8", "replace")
        component = resource.split("/")[0]
        if not component:
            return
        if not component in root.keys():
            root[component] = {}
        resource = "/".join(resource.split("/")[1:])
        _put_into_dict(root[component], resource=resource)

    for entry in data:
        (resource, code, ext, response_size) = (entry[3], entry[5], entry[4], entry[6])

        if banned_extensions and is_banned(resource):
            continue

        if unbanned_extensions and not is_ubanned(resource):
            continue

        if size_filter and size_is_banned(response_size):
            continue
        if code in response_codes:
            if ext:
                resource = resource + ext
            _put_into_dict(tree["/"], resource)

    print(tr(tree))


if __name__ == "__main__":
    if args.tree:
        ftree()
    else:
        table()
